@@ -50,77 +50,72 @@ class PathManager:
 class Meaurement:
     def __init__(self, path_mgt):
         self.path_mgt = path_mgt
-        self.model_name = None
-        self.use_openAI = False
-        self.model = None
-        self.input_df = None
-        self.embeddings = None
-        self.col_names = None
-        self.doc_col_name = None
-        self.doc_id_col_name = None
-        self.n_dims = None
-        self.n_scales = None
-        self.dim_embeddings = {}
-        self.dim_queries = {}
-        self.scale_embeddings = {}
-        self.scale_definitions = {}
 
-    def read_csv_cols(self, file_obj):
+    def read_csv_cols(self, file_obj, measurement_state):
         if file_obj is not None:
-            self.input_df = pd.read_csv(file_obj[0].name, encoding="ISO-8859-1")
-            self.col_names = self.input_df.columns.tolist()
-            print(f"Columns: {self.col_names}")
+            measurement_state["input_df"] = pd.read_csv(
+                file_obj[0].name, encoding="ISO-8859-1"
+            )
+            measurement_state["col_names"] = measurement_state[
+                "input_df"
+            ].columns.tolist()
+            print(f"Columns: {measurement_state['col_names']}")
         return gr.Dropdown.update(
-            choices=self.col_names, interactive=True
-        ), gr.Dropdown.update(choices=self.col_names, interactive=True)
+            choices=measurement_state["col_names"], interactive=True
+        ), gr.Dropdown.update(choices=measurement_state["col_names"], interactive=True)
 
-    def read_input_embedding(self, file_obj):
-        self.embeddings = np.load(file_obj[0].name)
-        if self.embeddings.shape[0] != len(self.input_df):
-            msg = f"Error: The number of rows in the input data and the number of rows in the embedding matrix do not match! The shape of the embeddings is {self.embeddings.shape}, and the number of rows in the input data is {len(self.input_df)}. "
+    def read_input_embedding(self, file_obj, measurement_state):
+        measurement_state["embeddings"] = np.load(file_obj[0].name)
+        if measurement_state["embeddings"].shape[0] != len(
+            measurement_state["input_df"]
+        ):
+            msg = f"Error: The number of rows in the input data and the number of rows in the embedding matrix do not match! The shape of the embeddings is {measurement_state['embeddings'].shape}, and the number of rows in the input data is {len(measurement_state['input_df'])}. "
         else:
-            msg = f"Precomputed embedding uploaded! \nThe shape of the embeddings is {self.embeddings.shape}. Proceed to the next tab to define the dimensions."
+            msg = f"Precomputed embedding uploaded! \nThe shape of the embeddings is {measurement_state['embeddings'].shape}. Proceed to the next tab to define the dimensions."
         return gr.Textbox.update(value=msg), gr.Button.update(visible=False)
 
-    def set_doc_col(self, doc_col_name, doc_id_col_name):
+    def set_doc_col(self, doc_col_name, doc_id_col_name, measurement_state):
         print(f"Setting doc_col_name to {doc_col_name}")
-        if self.col_names is not None:
-            if doc_col_name in self.col_names:
-                self.doc_col_name = doc_col_name
+        if measurement_state["input_df"].columns.tolist() is not None:
+            if doc_col_name in measurement_state["input_df"].columns.tolist():
+                measurement_state["doc_col_name"] = doc_col_name
 
         print(f"Setting doc_id_col_name to {doc_id_col_name}")
-        if self.col_names is not None:
-            if doc_id_col_name in self.col_names:
-                self.doc_id_col_name = doc_id_col_name
+        if measurement_state["input_df"].columns.tolist() is not None:
+            if doc_id_col_name in measurement_state["input_df"].columns.tolist():
+                measurement_state["doc_id_col_name"] = doc_id_col_name
         return gr.Textbox.update(value="Column names set!")
 
-    def set_sbert(self, model_name):
+    def set_sbert(self, model_name, measurement_state):
         print(f"Setting SBERT model to {model_name}")
-        self.model_name = model_name
-        self.tokenizer = AutoTokenizer.from_pretrained(
+        measurement_state["model_name"] = model_name
+        measurement_state["tokenizer"] = AutoTokenizer.from_pretrained(
             f"sentence-transformers/{model_name}"
         )
-        self.model = AutoModel.from_pretrained(f"sentence-transformers/{model_name}")
-        self.model.eval()
-        self.model.to(device)
+        measurement_state["model"] = AutoModel.from_pretrained(
+            f"sentence-transformers/{model_name}"
+        )
+        measurement_state["model"].eval()
+        measurement_state["model"].to(device)
 
-    def set_openai_api_key(self, api_key):
+    def set_openai_api_key(self, api_key, measurement_state):
         print("Setting OpenAI API key")
         if api_key != "":
             try:
-                openai.api_key = api_key
+                measurement_state["openai_api_key"] = api_key
+                print("Using API key from textbox.")
             except Exception as e:
                 print(e)
                 return gr.Textbox.update(
-                    value="Invalid API key. Please try again.", visible=True
+                    value="⚠️ Invalid API key. Please try again.", visible=True
                 )
         else:
             print("Using environment variable (OPENAI_API_KEY) for API key.")
             try:
-                openai.api_key = os.environ["OPENAI_API_KEY"]
+                measurement_state["openai_api_key"] = os.environ["OPENAI_API_KEY"]
             except Exception as e:
                 return gr.Textbox.update(
-                    value="Invalid API key in OPENAI_API_KEY environment variable. Please try again.",
+                    value="⚠️ Invalid API key in OPENAI_API_KEY environment variable. Please try again.",
                     visible=True,
                 )
 
@@ -146,14 +141,15 @@ class Meaurement:
         embedding_model_dropdown,
         sbert_model_textbox,
         openai_api_key,
+        measurement_state,
     ):
         print(f"Setting embedding model to {embedding_model_dropdown}")
         if embedding_model_dropdown.startswith("Sentence Transformers"):
-            self.set_sbert(sbert_model_textbox)
-            self.use_openAI = False
+            self.set_sbert(sbert_model_textbox, measurement_state)
+            measurement_state["use_openAI"] = False
         if embedding_model_dropdown.startswith("OpenAI"):
-            self.set_openai_api_key(openai_api_key)
-            self.use_openAI = True
+            self.set_openai_api_key(openai_api_key, measurement_state)
+            measurement_state["use_openAI"] = True
         return gr.Button.update(value="Embedding model set!")
 
     def reset_set_emb_btn(self):
@@ -165,24 +161,26 @@ class Meaurement:
         if embedding_option == "Embed Documents":
             return gr.File.update(visible=False), gr.Button.update(visible=True)
 
-    def embed_texts(self, sentences, progress, batch_size=8):
+    @classmethod
+    def embed_texts(cls, sentences, progress, measurement_state):
         """Use huggingface transformers to embed the text_col"""
+        batch_size = 8
         sentences = [str(x) for x in sentences]
-        if self.use_openAI is False:
+        if measurement_state["use_openAI"] is False:
             # use sentence_transformers to embed the text_col
             print("Embedding Using Sentence Transformers")
             # use batching
             with torch.no_grad():
-                self.model.eval()
+                measurement_state["model"].eval()
                 sentence_embeddings = []
                 for i in progress.tqdm(
                     range(0, len(sentences), batch_size), unit="batch (batch size=8)"
                 ):
                     batch = sentences[i : i + batch_size]
-                    encoded_input = self.tokenizer(
+                    encoded_input = measurement_state["tokenizer"](
                         batch, padding=True, truncation=True, return_tensors="pt"
                     ).to(device)
-                    model_output = self.model(**encoded_input)
+                    model_output = measurement_state["model"](**encoded_input)
                     batch_sentence_embeddings = util_funcs.mean_pooling(
                         model_output, encoded_input["attention_mask"]
                     )
@@ -200,7 +198,9 @@ class Meaurement:
             sentence_embeddings = []
             for sent in progress.tqdm(sentences, unit="docs"):
                 response = openai.Embedding.create(
-                    input=sent, model="text-embedding-ada-002"
+                    input=sent,
+                    model="text-embedding-ada-002",
+                    api_key=measurement_state["openai_api_key"],
                 )
                 embeddings = response["data"][0]["embedding"]
                 sentence_embeddings.append(embeddings)
@@ -212,59 +212,40 @@ class Meaurement:
 
         return sentence_embeddings
 
-    def embed_df(self, progress=gr.Progress()):
+    def embed_df(self, measurement_state, progress=gr.Progress()):
         try:
-            self.embeddings = self.embed_texts(
-                self.input_df[self.doc_col_name], progress=progress
+            measurement_state["embeddings"] = self.embed_texts(
+                measurement_state["input_df"][measurement_state["doc_col_name"]],
+                progress=progress,
+                measurement_state=measurement_state,
             )
-            np.save(Path(self.path_mgt.out_dir, "embeddings.npy"), self.embeddings)
-            if (self.doc_col_name is not None) and (
+            np.save(
+                Path(self.path_mgt.out_dir, "embeddings.npy"),
+                measurement_state["embeddings"],
+            )
+            if (measurement_state["doc_col_name"] is not None) and (
                 Path(self.path_mgt.out_dir, "embeddings.npy").exists()
             ):
                 return (
                     gr.Textbox.update(
-                        value=f"Embedding completed! \nThe shape of the embeddings is {self.embeddings.shape}. You can download and save the embeddings below. Proceed to the next tab to define the dimensions."
+                        value=f"Embedding completed! \nThe shape of the embeddings is {measurement_state['embeddings'].shape}. You can download and save the embeddings below. Proceed to the next tab to define the dimensions."
                     ),
                     gr.File.update(visible=True),
                     Path(self.path_mgt.out_dir, "embeddings.npy"),
                 )
         except Exception as e:
             return (
-                gr.Textbox.update(value=f"Embedding failed! \nError message: {e}"),
+                gr.Textbox.update(
+                    value=f"⚠️ Embedding failed! Make sure that you have clicked Confirm Column Selections and Set Embedding Model. \nError message: {e}"
+                ),
                 gr.File.update(visible=False),
                 None,
             )
 
-    def init_query_boxes(self, n_dims):
-        # ! deprecated
-        # toggle visibility of query boxes, dim name, search buttons, and results
-        update_boxes = []
-        self.n_dims = n_dims
-        for i in range(10):
-            if i < n_dims:
-                update_boxes.extend(
-                    [
-                        gr.Textbox.update(visible=True),
-                        gr.Textbox.update(visible=True),
-                        gr.Button.update(visible=True),
-                        gr.Textbox.update(visible=True),
-                    ]
-                )
-            else:
-                update_boxes.extend(
-                    [
-                        gr.Textbox.update(visible=False),
-                        gr.Textbox.update(visible=False),
-                        gr.Button.update(visible=False),
-                        gr.Textbox.update(visible=False),
-                    ]
-                )
-        return update_boxes
-
-    def toggle_row_vis(self, n_rows):
+    def toggle_row_vis(self, n_rows, measurement_state):
         # toggle visbility of rows (dimensions tab)
         update_rows = []
-        self.n_dims = n_rows
+        measurement_state["n_dims"] = n_rows
         for i in range(10):
             if i < n_rows:
                 update_rows.append(gr.Row.update(visible=True))
@@ -272,10 +253,10 @@ class Meaurement:
                 update_rows.append(gr.Row.update(visible=False))
         return update_rows
 
-    def toggle_row_vis_scales(self, n_rows):
+    def toggle_row_vis_scales(self, n_rows, measurement_state):
         # toggle visbility of rows (scales tab)
         update_rows = []
-        self.n_scales = n_rows
+        measurement_state["n_scales"] = n_rows
         for i in range(10):
             if i < n_rows:
                 update_rows.append(gr.Row.update(visible=True))
@@ -283,26 +264,36 @@ class Meaurement:
                 update_rows.append(gr.Row.update(visible=False))
         return update_rows
 
-    def semantic_search(self, query, n_results, progress=gr.Progress()):
+    def semantic_search(
+        self, query, n_results, measurement_state, progress=gr.Progress()
+    ):
         # search for the most similar documents
         result_str = ""
         queries = query.split("\n")
         # remove empty queries or queries with only spaces
         queries = [q for q in queries if q.strip() != ""]
+        if (len(queries) == 0) or (queries is None) or all(q == "" for q in queries):
+            return gr.Textbox.update(value="No query provided!")
         try:
-            query_embedding = self.embed_texts(queries, progress=progress)
+            query_embedding = self.embed_texts(
+                queries, progress=progress, measurement_state=measurement_state
+            )
             mean_vect_query = query_embedding.mean(axis=0)
             hits = util_funcs.semantic_search(
                 mean_vect_query,
-                self.embeddings,
+                measurement_state["embeddings"],
                 score_function=util_funcs.dot_score,
                 top_k=n_results,
             )
             hit_ids = [h["corpus_id"] for h in hits[0]]
             hit_scores = [h["score"] for h in hits[0]]
-            examples = itemgetter(*hit_ids)(self.input_df[self.doc_col_name])
+            examples = itemgetter(*hit_ids)(
+                measurement_state["input_df"][measurement_state["doc_col_name"]]
+            )
             examples = list(examples)
-            example_doc_ids = itemgetter(*hit_ids)(self.input_df[self.doc_id_col_name])
+            example_doc_ids = itemgetter(*hit_ids)(
+                measurement_state["input_df"][measurement_state["doc_id_col_name"]]
+            )
             example_doc_ids = list(example_doc_ids)
             assert len(hit_ids) == len(hit_scores) == len(examples)
             for i in range(len(hit_ids)):
@@ -313,18 +304,18 @@ class Meaurement:
                 result_str += "------------------------ \n"
         except Exception as e:
             result_str = (
-                f"Error. Make sure the query is not empty. \nError message: {e}"
+                f"⚠️ Error. Make sure the query is not empty. \nError message: {e}"
             )
         return gr.Textbox.update(value=result_str)
 
-    def save_dims(self, progress=gr.Progress(), *dims_boxes):
+    def save_dims(self, measurement_state, progress=gr.Progress(), *dims_boxes):
         try:
             # first half of dims_boxes are dim names, second half are dim queries
             all_dim_names = dims_boxes[: len(dims_boxes) // 2]
             all_dim_queries = dims_boxes[len(dims_boxes) // 2 :]
-            self.dim_embeddings = {}
-            self.dim_queries = {}
-            for i in range(self.n_dims):
+            measurement_state["dim_embeddings"] = {}
+            measurement_state["dim_queries"] = {}
+            for i in range(measurement_state["n_dims"]):
                 if all_dim_names[i].strip() != "":
                     dim_name = all_dim_names[i].strip()
                 else:  # if no name is given, use default name
@@ -332,39 +323,53 @@ class Meaurement:
                 dim_queries = all_dim_queries[i].split("\n")  # dim queries
                 # remove empty queries or queries with only spaces
                 dim_queries = [q for q in dim_queries if q.strip() != ""]
-                dim_embedding = self.embed_texts(dim_queries, progress=progress)
+                dim_embedding = self.embed_texts(
+                    dim_queries, progress=progress, measurement_state=measurement_state
+                )
                 mean_vect_dim = dim_embedding.mean(axis=0)
-                self.dim_embeddings[dim_name] = mean_vect_dim
-                self.dim_queries[dim_name] = dim_queries
-            # save self.dim_queries to json file
+                measurement_state["dim_embeddings"][dim_name] = mean_vect_dim
+                measurement_state["dim_queries"][dim_name] = dim_queries
+            # save measurement_state['dim_queries'] to json file
             with open(Path(self.path_mgt.out_dir, "dimensions_queries.json"), "w") as f:
-                json.dump(self.dim_queries, f)
+                json.dump(measurement_state["dim_queries"], f)
             # update both positive and negative scale selectors
             return (
-                [gr.Dropdown.update(choices=list(self.dim_queries.keys()))] * 20
+                [
+                    gr.Dropdown.update(
+                        choices=list(measurement_state["dim_queries"].keys())
+                    )
+                ]
+                * 20
                 + [
                     gr.Textbox.update(
                         visible=True,
-                        value=f"Dimensions saved: {list(self.dim_queries.keys())}. You can download the json file below to keep a record of the final queries. Proceed to the next tab to define scales.",
+                        value=f"Dimensions saved: {list(measurement_state['dim_queries'].keys())}. You can download the json file below to keep a record of the final queries. Proceed to the next tab to define scales.",
                     )
                 ]
                 + [gr.File.update(visible=True)]
                 + [Path(self.path_mgt.out_dir, "dimensions_queries.json")]
+                + [gr.Markdown.update(visible=False)]
             )
         except Exception as e:
             return (
-                [gr.Dropdown.update(choices=list(self.dim_queries.keys()))] * 20
+                [
+                    gr.Dropdown.update(
+                        choices=list(measurement_state["dim_queries"].keys())
+                    )
+                ]
+                * 20
                 + [
                     gr.Textbox.update(
                         visible=True,
-                        value=f"Please check that you have filled in all the dimension names and queries. \n  Error message: {e}.",
+                        value=f"⚠️ Please check that you have filled in all the dimension names and queries. \n  Error message: {e}.",
                     )
                 ]
                 + [gr.File.update(visible=False)]
                 + [None]
+                + [gr.Markdown.update(visible=True)]
             )
 
-    def save_scales(self, *scale_boxes):
+    def save_scales(self, measurement_state, *scale_boxes):
         # first 1/3 of scales boxes are names, second 1/3 are positive scales, third 1/3 are negative scales
         try:
             all_scale_names = scale_boxes[: len(scale_boxes) // 3]
@@ -372,10 +377,10 @@ class Meaurement:
                 len(scale_boxes) // 3 : 2 * len(scale_boxes) // 3
             ]
             all_neg_scales = scale_boxes[2 * len(scale_boxes) // 3 :]
-            self.scale_embeddings = {}
-            self.scale_definitions = {}
+            measurement_state["scale_embeddings"] = {}
+            measurement_state["scale_definitions"] = {}
             # all scale embeddings are average of positive scales subtracted by average of negative scales
-            for i in range(self.n_scales):
+            for i in range(measurement_state["n_scales"]):
                 if all_scale_names[i].strip() != "":
                     scale_name = all_scale_names[i].strip()
                 else:  # if no name is given, use default name
@@ -394,7 +399,7 @@ class Meaurement:
                     },
                 }
                 """
-                self.scale_definitions[scale_name] = {
+                measurement_state["scale_definitions"][scale_name] = {
                     "Positive": all_pos_scales[i],
                     "Negative": all_neg_scales[i],
                 }
@@ -402,10 +407,14 @@ class Meaurement:
                 scale_embedding_neg = []
                 if len(all_pos_scales[i]) > 0:
                     for pos_scale in all_pos_scales[i]:
-                        scale_embedding_pos.append(self.dim_embeddings[pos_scale])
+                        scale_embedding_pos.append(
+                            measurement_state["dim_embeddings"][pos_scale]
+                        )
                 if len(all_neg_scales[i]) > 0:
                     for neg_scale in all_neg_scales[i]:
-                        scale_embedding_neg.append(self.dim_embeddings[neg_scale])
+                        scale_embedding_neg.append(
+                            measurement_state["dim_embeddings"][neg_scale]
+                        )
                 if (len(scale_embedding_pos) > 0) & (len(scale_embedding_neg) > 0):
                     # take averages and difference
                     scale_embedding = np.stack(scale_embedding_pos).mean(
@@ -416,22 +425,26 @@ class Meaurement:
                 elif len(scale_embedding_neg) > 0:  # only negative scales
                     scale_embedding = -np.stack(scale_embedding_neg).mean(axis=0)
                 else:  # no scales
-                    pass
-                self.scale_embeddings[scale_name] = scale_embedding
+                    # throw exception
+                    raise Exception(
+                        f"Check that you have filled in all the scale names and at least one positive or negative scale for scale {scale_name}."
+                    )
+                measurement_state["scale_embeddings"][scale_name] = scale_embedding
 
-            # save self.scales to json file
+            # save measurement_state['scales'] to json file
             with open(Path(self.path_mgt.out_dir, "scale_definitions.json"), "w") as f:
-                json.dump(self.scale_definitions, f)
+                json.dump(measurement_state["scale_definitions"], f)
 
             return (
                 [
                     gr.Textbox.update(
                         visible=True,
-                        value=f"Scales saved: {list(self.scale_definitions.keys())}. You can download the json file below to keep a record of the scale definitions. Proceed to the next tab to measure using semantic projection.",
+                        value=f"Scales saved: {list(measurement_state['scale_definitions'].keys())}. You can download the json file below to keep a record of the scale definitions. Proceed to the next tab to measure using semantic projection.",
                     )
                 ]
                 + [gr.File.update(visible=True)]
                 + [Path(self.path_mgt.out_dir, "scale_definitions.json")]
+                + [gr.Markdown.update(visible=False)]
             )
         except Exception as e:
             # return error message in textbox update
@@ -439,19 +452,21 @@ class Meaurement:
                 [
                     gr.Textbox.update(
                         visible=True,
-                        value=f"Error: Please check your scale definitions. Make sure that you have clicked the Embed Queries and Save Dimensions button in Tab 2, and there is no empty scales or dimensions. \nError message: {e}",
+                        value=f"⚠️ Error: Please check your scale definitions. Make sure that you have clicked the Embed Queries and Save Dimensions button in Tab 2, and there is no empty scales or dimensions. \nError message: {e}",
                     )
                 ]
                 + [gr.File.update(visible=False)]
                 + [None]
+                + [gr.Markdown.update(visible=True)]
             )
 
-    def measure_docs(self, whitening: str):
+    def measure_docs(self, whitening: str, measurement_state):
         scale_measures = []
-        for scale in list(self.scale_definitions.keys()):
+        for scale in list(measurement_state["scale_definitions"].keys()):
             measures = (
                 util_funcs.pytorch_cos_sim(
-                    self.embeddings, self.scale_embeddings[scale]
+                    measurement_state["embeddings"],
+                    measurement_state["scale_embeddings"][scale],
                 )
                 .cpu()
                 .numpy()
@@ -468,10 +483,13 @@ class Meaurement:
         scale_measures = scale_measures.round(4)
         # output as csv to File
         scale_measures = pd.DataFrame(scale_measures)
-        scale_measures.columns = list(self.scale_definitions.keys())
-        scale_measures[self.doc_id_col_name] = self.input_df[self.doc_id_col_name]
+        scale_measures.columns = list(measurement_state["scale_definitions"].keys())
+        scale_measures[measurement_state["doc_id_col_name"]] = measurement_state[
+            "input_df"
+        ][measurement_state["doc_id_col_name"]]
         scale_measures = scale_measures[
-            [self.doc_id_col_name] + list(self.scale_definitions.keys())
+            [measurement_state["doc_id_col_name"]]
+            + list(measurement_state["scale_definitions"].keys())
         ]
         Path("measure_output").mkdir(parents=True, exist_ok=True)
         Path(self.path_mgt.out_dir, "measurement_output.csv")
@@ -487,16 +505,18 @@ class Meaurement:
             Path(self.path_mgt.out_dir, "measurement_output.csv"),
         )
 
-    def load_example_dataset(self):
+    def load_example_dataset(self, measurement_state):
         # tab 1
-        self.input_df = pd.read_csv(
+        measurement_state["input_df"] = pd.read_csv(
             Path(self.path_mgt.sample_data_dir, "sample_text.csv")
         )
-        self.embeddings = np.load(Path(self.path_mgt.sample_data_dir, "sample_emb.npy"))
-        self.set_sbert("all-MiniLM-L6-v2")
-        self.doc_col_name = "text"
-        self.doc_id_col_name = "doc_id"
-        self.use_openAI = False
+        measurement_state["embeddings"] = np.load(
+            Path(self.path_mgt.sample_data_dir, "sample_emb.npy")
+        )
+        self.set_sbert("all-MiniLM-L6-v2", measurement_state)
+        measurement_state["doc_col_name"] = "text"
+        measurement_state["doc_id_col_name"] = "doc_id"
+        measurement_state["use_openAI"] = False
 
         # tab 2
         all_dim_name_boxes_updates = []
@@ -577,6 +597,7 @@ def run_gui(
 
     with gr.Blocks(title="SPAR") as demo:
         m = Meaurement(path_mgt=path_mgt)
+        state = gr.State({})
         gr.Markdown(
             "### SPAR: Semantic Projection with Active Retrieval (Research Preview)"
         )
@@ -622,7 +643,7 @@ def run_gui(
         with gr.Tab("1. Upload File and Embed"):
             # read csv file and select columns
             gr.Markdown(
-                value=" 📖 Upload a **CSV file** that contains at least 2 columns: (1) the documents to be measured, and (2) the document IDs. Alternatively, you can click the Load Example Dataset and Scales button on top to explore the tool with an included sample dataset (4000 Facebook posts) and pre-defined Competing Values Framework (CVF) dimension and scales.",
+                value=" 📖 Upload a **CSV file** that contains at least 2 columns: (1) the documents to be measured, and (2) the document IDs. Alternatively, you can click the Load Example Dataset and Scales button on top to explore the tool with an included sample dataset (2000 Facebook posts) and pre-defined Competing Values Framework (CVF) dimension and scales.",
                 label="",
             )
             with gr.Row(variant="panel"):
@@ -646,13 +667,13 @@ def run_gui(
                 doc_id_col_btn = gr.Button(value="Confirm Column Selections")
                 doc_id_col_btn.click(
                     fn=m.set_doc_col,
-                    inputs=[doc_col_selector, doc_id_col_selector],
+                    inputs=[doc_col_selector, doc_id_col_selector, state],
                     outputs=doc_id_col_btn,
                 )
 
             gr.Markdown(
                 value=""" 📖 Select an embedding method. You can use Sentence Transformers to embed the text locally; in this case you can use the default model name (all-MiniLM-L6-v2) or [any other models](https://www.sbert.net/docs/pretrained_models.html).
-                You can also use the [OpenAI API](https://platform.openai.com/docs/guides/embeddings/what-are-embeddings) to embed (text-embedding-ada-002 model). You can get an API key [here](https://platform.openai.com/account/api-keys). """,
+                Alternatively, you can use the [OpenAI API](https://platform.openai.com/docs/guides/embeddings/what-are-embeddings) to embed (text-embedding-ada-002 model). You can get an API key [here](https://platform.openai.com/account/api-keys). """,
             )
 
             with gr.Row(variant="panel"):
@@ -688,6 +709,7 @@ def run_gui(
                         embedding_model_dropdown,
                         sbert_model_textbox,
                         openai_api_key,
+                        state,
                     ],
                     outputs=set_emb_btn,
                 )
@@ -723,13 +745,13 @@ def run_gui(
 
             input_file.change(
                 fn=m.read_csv_cols,
-                inputs=input_file,
+                inputs=[input_file, state],
                 outputs=[doc_col_selector, doc_id_col_selector],
             )
 
             input_embedding_uploader.change(
                 fn=m.read_input_embedding,
-                inputs=input_embedding_uploader,
+                inputs=[input_embedding_uploader, state],
                 outputs=[emb_result_txtbox, embed_btn],
             )
             embedding_model_dropdown.change(
@@ -746,11 +768,12 @@ def run_gui(
             # embedding
             openai_api_key.change(
                 fn=m.set_openai_api_key,
-                inputs=openai_api_key,
+                inputs=[openai_api_key, state],
                 outputs=[emb_result_txtbox],
             )
             embed_btn.click(
                 fn=m.embed_df,
+                inputs=state,
                 outputs=[emb_result_txtbox, emb_results_file, emb_results_file],
             )
         with gr.Tab("2. Define Dimensions and Semantic Search"):
@@ -776,7 +799,7 @@ def run_gui(
             )
 
             gr.Markdown(
-                value=" 📖 Enter the names of the dimensions and seed search queries. Then click the Search Dimension button to search for relevant documents in the corpus. Copy, add, and edit the relevant documents to query box to conduct next round of search. Each query should be in its own line. ",
+                value=" 📖 Enter the names of the dimensions and seed search queries. Then click 'Search Dimension' to search for relevant documents in the corpus. \nCopy & paste the relevant documents to query box (and/or edit) to conduct next round of search. You can use multiple queries in each dimension, with each query entered on its own line. ",
                 label="",
             )
 
@@ -854,29 +877,30 @@ def run_gui(
                         )
 
             n_dim_slider.change(
-                fn=m.toggle_row_vis, inputs=n_dim_slider, outputs=all_rows_dims
+                fn=m.toggle_row_vis, inputs=[n_dim_slider, state], outputs=all_rows_dims
             )
             for box in all_search_query_boxes:
-                box.change(fn=m.toggle_row_vis, inputs=n_dim_slider)
+                box.change(fn=m.toggle_row_vis, inputs=[n_dim_slider, state])
 
             for dim_i, btn in enumerate(all_search_btns):
                 btn.click(
                     fn=m.semantic_search,
-                    inputs=[
-                        all_search_query_boxes[dim_i],
-                        n_results_slider,
-                    ],
+                    inputs=[all_search_query_boxes[dim_i], n_results_slider, state],
                     outputs=all_search_results[dim_i],
                 )
             gr.Markdown(
-                value=" 📖 After defining the dimensions and final queries, click the Embed Queries and Save Dimensions button to embed the queries and save the dimension definitions. Each dimension must contain at least one query.",
+                value=" 📖 After defining the dimensions with final context-specific queries, click 'Embed Queries and Save Dimensions' button below to embed the queries and save the definitions. Each dimension must contain at least one final query.",
                 label="",
             )
             save_dim_button = gr.Button("Embed Queries and Save Dimensions")
             dim_define_results = gr.Textbox(visible=False, label="")
             dimension_def_file_download = gr.File(visible=False)
 
-        with gr.Tab("3. Define Scales"):
+        with gr.Tab("3. Define Scales", visible=False):
+            tab2_warn = gr.Markdown(
+                value="<span style='color:red'>⚠️ Embed Queries and Save Dimensions button in Tab 2 must be clicked before defining scales.</span>",
+                visible=True,
+            )
             gr.Markdown(
                 value=" 📖 Move the sliders to set the number of scales.",
                 label="",
@@ -947,37 +971,46 @@ def run_gui(
                                 visible=True,
                             )
                         )
-            save_dim_button.click(
-                fn=m.save_dims,
-                inputs=all_dim_name_boxes + all_search_query_boxes,
-                outputs=all_scale_pos_selector
-                + all_scale_neg_selector
-                + [dim_define_results]
-                + [dimension_def_file_download] * 2,
-            )
             n_scale_slider.change(
                 fn=m.toggle_row_vis_scales,
-                inputs=n_scale_slider,
+                inputs=[n_scale_slider, state],
                 outputs=all_rows_scale,
             )
             for box in all_scale_name_boxes:
-                box.change(fn=m.toggle_row_vis_scales, inputs=n_scale_slider)
+                box.change(fn=m.toggle_row_vis_scales, inputs=[n_scale_slider, state])
             for box in all_scale_pos_selector:
-                box.change(fn=m.toggle_row_vis_scales, inputs=n_scale_slider)
+                box.change(fn=m.toggle_row_vis_scales, inputs=[n_scale_slider, state])
             for box in all_scale_neg_selector:
-                box.change(fn=m.toggle_row_vis_scales, inputs=n_scale_slider)
+                box.change(fn=m.toggle_row_vis_scales, inputs=[n_scale_slider, state])
             save_scale_button = gr.Button("Save Scales")
             scale_define_results = gr.Textbox(visible=False, label="")
             scale_def_file_download = gr.File(visible=False)
 
-            save_scale_button.click(
-                fn=m.save_scales,
-                inputs=all_scale_name_boxes
-                + all_scale_pos_selector
-                + all_scale_neg_selector,
-                outputs=[scale_define_results] + [scale_def_file_download] * 2,
+            save_dim_button.click(
+                fn=m.save_dims,
+                inputs=[state] + all_dim_name_boxes + all_search_query_boxes,
+                outputs=all_scale_pos_selector
+                + all_scale_neg_selector
+                + [dim_define_results]
+                + [dimension_def_file_download] * 2
+                + [tab2_warn],
             )
         with gr.Tab("4. Measurement"):
+            tab3_warn = gr.Markdown(
+                value="<span style='color:red'>⚠️ Save Scales button in Tab 3 must be clicked before measurement.</span>",
+                visible=True,
+            )
+            save_scale_button.click(
+                fn=m.save_scales,
+                inputs=[state]
+                + all_scale_name_boxes
+                + all_scale_pos_selector
+                + all_scale_neg_selector,
+                outputs=[scale_define_results]
+                + [scale_def_file_download] * 2
+                + [tab3_warn],
+            )
+
             gr.Markdown(
                 value=" 📖 Click the Measure Documents Using Semantic Projection button to score each document. The output file will contain the document ID and the scores for each scale.",
                 label="",
@@ -995,11 +1028,12 @@ def run_gui(
             measure_results_file = gr.File(visible=False)
             measure_button.click(
                 fn=m.measure_docs,
-                inputs=whitening_radio_btn,
+                inputs=[whitening_radio_btn, state],
                 outputs=[measure_result, measure_results_file, measure_results_file],
             )
         example_btn.click(
             fn=m.load_example_dataset,
+            inputs=state,
             outputs=[
                 example_row,
                 example_file_download,
@@ -1018,7 +1052,7 @@ def run_gui(
             + all_scale_neg_selector
             + [n_dim_slider, n_scale_slider],
         )
-    demo.queue(concurrency_count=3)
+    demo.queue(concurrency_count=2)
     if username is None or password is None:
         auth = None
     else:
